from .framework import (
    selenium_test,
    SeleniumTestCase
)


class HistoryPanelTestCase(SeleniumTestCase):

    ensure_registered = True

    @selenium_test
    def test_history_panel_landing_state(self):
        self.assert_initial_history_panel_state_correct()

        tag_icon_selector = self.navigation.history_panel.selectors.tag_icon
        annotation_icon_selector = self.navigation.history_panel.selectors.annotation_icon

        self.wait_for_visible(tag_icon_selector)
        self.wait_for_visible(annotation_icon_selector)

        name_element = self.history_panel_name_element()
        self.assert_tooltip_text(name_element, self.navigation.history_panel.text.tooltip_name)

    @selenium_test
    def test_history_panel_rename(self):
        editable_text_input_element = self.history_panel_click_to_rename()
        editable_text_input_element.send_keys("New History Name")
        self.send_enter(editable_text_input_element)

        assert "New History Name" in self.history_panel_name()

    @selenium_test
    def test_history_rename_confirm_with_click(self):
        editable_text_input_element = self.history_panel_click_to_rename()
        editable_text_input_element.send_keys("New History Name")
        self.click_center()
        self.assert_absent(self.navigation.history_panel.selectors.name_edit_input)
        assert "New History Name" in self.history_panel_name()

    @selenium_test
    def test_history_rename_cancel_with_escape(self):
        editable_text_input_element = self.history_panel_click_to_rename()
        editable_text_input_element.send_keys("New History Name")
        self.send_escape(editable_text_input_element)
        self.assert_absent(self.navigation.history_panel.selectors.name_edit_input)
        assert "New History Name" not in self.history_panel_name()

    @selenium_test
    def test_history_tags_and_annotations_buttons(self):
        tag_icon_selector = self.navigation.history_panel.selectors.tag_icon
        annotation_icon_selector = self.navigation.history_panel.selectors.annotation_icon

        tag_area_selector = self.navigation.history_panel.selectors.tag_area
        annotation_area_selector = self.navigation.history_panel.selectors.annotation_area

        tag_icon = self.wait_for_clickable(tag_icon_selector)
        annon_icon = self.wait_for_clickable(annotation_icon_selector)

        self.assert_absent_or_hidden(tag_area_selector)
        self.assert_absent_or_hidden(annotation_area_selector)

        tag_icon.click()

        self.wait_for_visible(tag_area_selector)
        self.assert_absent_or_hidden(annotation_area_selector)

        tag_icon.click()
        self.sleep_for(self.wait_types.UX_TRANSITION)
        annon_icon.click()

        self.wait_for_visible(annotation_area_selector)
        self.assert_absent_or_hidden(tag_area_selector)

        annon_icon.click()
        self.sleep_for(self.wait_types.UX_TRANSITION)

        self.assert_absent_or_hidden(tag_area_selector)
        self.assert_absent_or_hidden(annotation_area_selector)

    @selenium_test
    def test_history_panel_annotations_change(self):

        def assert_current_annotation(expected, error_message="History annotation",
                                      is_equal=True):
            current_annotation = self.components.history_panel.annotation_editable_text.wait_for_visible()
            error_message += " given: [%s] expected [%s] "
            if is_equal:
                assert current_annotation.text == expected, error_message % (
                    current_annotation.text, expected)
            else:
                assert current_annotation.text != expected, error_message % (
                    current_annotation.text, expected)

        def set_random_annotation(clear_text=True):
            random_annotation = self._get_random_name(prefix="arbitrary_annotation_")
            self.set_history_annotation(random_annotation, clear_text)
            return random_annotation

        # assert that annotation wasn't set before
        self.components.history_panel.annotation_area.assert_absent_or_hidden()

        # assign annotation random text
        initial_annotation = set_random_annotation()
        assert_current_annotation(initial_annotation)

        # change annotation text
        changed_annotation = set_random_annotation()

        assert_current_annotation(initial_annotation, error_message="History annotation was not changed!",
                                  is_equal=False)
        assert_current_annotation(changed_annotation,
                                  error_message="History annotation was changed, but annotation text is wrong!",
                                  is_equal=True)

    @selenium_test
    def test_history_panel_tags_change(self):

        def add_tags(size):
            history_panel_tags = list()

            for i in range(size):
                history_panel_tags.append(self._get_random_name(prefix="arbitrary_tag_%s_") % i)

            self.history_panel_add_tags(history_panel_tags)
            return history_panel_tags

        def assert_current_tags(expected_tags):
            current_tags = self.components.history_panel.tags
            current_tags.wait_for_visible()
            assert [tag.text for tag in
                    current_tags.all()] == expected_tags, "tags [{}] are not the same as expected [{}]".format(current_tags, expected_tags)

        def clear_tags(expected_tags_size):

            close_tag_buttons = self.components.history_panel.tag_close_btn.all()
            current_tags_size = len(close_tag_buttons)

            assert expected_tags_size == current_tags_size, "there are more tags than expected! current {}, expected {}".format(
                current_tags_size, expected_tags_size)
            for close_btn in reversed(close_tag_buttons):
                close_btn.click()
                self.sleep_for(self.wait_types.UX_RENDER)

        tags_size = 5

        self.components.history_panel.tag_area.assert_absent_or_hidden()

        # add new tags to empty tags area
        tags = add_tags(tags_size)
        assert_current_tags(tags)

        # add more tags to non-empty tags area
        tags += add_tags(tags_size)
        assert_current_tags(tags)

        # delete all tags
        expected_tags_len = len(tags)
        clear_tags(expected_tags_len)
        self.components.history_panel.tags.assert_absent_or_hidden()

    @selenium_test
    def test_refresh_preserves_state(self):
        self.perform_upload(self.get_filename("1.txt"))
        self.wait_for_history()

        # Open the details, verify they are open and do a refresh.
        self.history_panel_ensure_showing_item_details(hid=1)
        self.history_panel_item_body_component(1, wait=True)
        self.history_panel_refresh_click()

        # After the refresh, verify the details are still open.
        self.sleep_for(self.wait_types.UX_TRANSITION)
        self.wait_for_selector_clickable(self.history_panel_item_selector(hid=1))
        assert self.history_panel_item_showing_details(hid=1)

        # Close the detailed display, refresh, and ensure they are still closed.
        self.history_panel_click_item_title(hid=1, wait=True)
        assert not self.history_panel_item_showing_details(hid=1)
        self.history_panel_refresh_click()
        self.sleep_for(self.wait_types.UX_TRANSITION)
        self.wait_for_selector_clickable(self.history_panel_item_selector(hid=1))
        assert not self.history_panel_item_showing_details(hid=1)



from .framework import (
    selenium_test,
    SeleniumTestCase
)


class CollectionBuildersTestCase(SeleniumTestCase):

    ensure_registered = True

    @selenium_test
    def test_build_list_simple(self):
        self.perform_upload(self.get_filename("1.fasta"))
        self.history_panel_wait_for_hid_ok(1)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_list)

        self.collection_builder_set_name("my cool list")
        self.screenshot("collection_builder_list")
        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(3)

    @selenium_test
    def test_build_list_and_hide_items(self):
        self.perform_upload(self.get_filename("1.fasta"))
        self.history_panel_wait_for_hid_ok(1)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_list)

        self.collection_builder_hide_originals()
        self.collection_builder_set_name("my cool list")

        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(3)
        self.history_panel_refresh_click()
        self.history_panel_wait_for_hid_hidden(1)

    @selenium_test
    def test_build_pair_simple(self):
        self.perform_upload(self.get_filename("1.tabular"))
        self.perform_upload(self.get_filename("2.tabular"))
        self._wait_for_hid_visible(1)
        self._wait_for_hid_visible(2)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_muli_operation_select_hid(2)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_pair)
        self.collection_builder_set_name("my awesome pair")
        self.screenshot("collection_builder_pair")
        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(5)

    @selenium_test
    def test_build_paired_list_simple(self):
        self.perform_upload(self.get_filename("1.tabular"))
        self.perform_upload(self.get_filename("2.tabular"))
        self._wait_for_hid_visible(1)
        self._wait_for_hid_visible(2)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_muli_operation_select_hid(2)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_list_pairs)

        self.collection_builder_clear_filters()
        self.collection_builder_click_paired_item("forward", 0)
        self.collection_builder_click_paired_item("reverse", 1)
        self.collection_builder_set_name("my awesome paired list")
        self.screenshot("collection_builder_paired_list")
        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(3)

    @selenium_test
    def test_build_paired_list_hide_original(self):
        self.perform_upload(self.get_filename("1.tabular"))
        self.perform_upload(self.get_filename("2.tabular"))
        self._wait_for_hid_visible(1)
        self._wait_for_hid_visible(2)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_muli_operation_select_hid(2)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_list_pairs)

        self.wait_for_and_click(self.navigation.collection_builders.selectors.clear_filters)

        forward_column = self.wait_for_visible(self.navigation.collection_builders.selectors.forward_datasets)
        first_datset_forward = forward_column.find_elements_by_css_selector("li")[0]
        first_datset_forward.click()

        reverse_column = self.wait_for_visible(self.navigation.collection_builders.selectors.reverse_datasets)
        second_dataset_reverse = reverse_column.find_elements_by_css_selector("li")[1]
        second_dataset_reverse.click()

        self.collection_builder_hide_originals()
        self.collection_builder_set_name("my awesome paired list")

        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(3)
        self.history_panel_wait_for_hid_hidden(1)
        self.history_panel_wait_for_hid_hidden(2)

    @selenium_test
    def test_build_simple_list_via_rules(self):
        self.perform_upload(self.get_filename("1.fasta"))
        self.history_panel_wait_for_hid_ok(1, allowed_force_refreshes=1)
        self.history_panel_multi_operations_show()
        self.history_panel_muli_operation_select_hid(1)
        self.history_panel_multi_operation_action_click(self.navigation.history_panel.multi_operations.labels.build_from_rules)

        self.collection_builder_set_name("my cool list")
        self.screenshot("collection_builder_rules_list")
        self.collection_builder_create()
        self.history_panel_wait_for_hid_ok(3, allowed_force_refreshes=1)

    def _wait_for_hid_visible(self, hid):
        self.history_panel_wait_for_hid_visible(hid, allowed_force_refreshes=1)



import unittest

from .framework import (
    selenium_test,
    SeleniumTestCase,
)


@unittest.skip
class LibraryToCollectionsTestCase(SeleniumTestCase):

    requires_admin = True

    @selenium_test
    def test_list_creation(self):
        self.admin_login()
        self.perform_upload(self.get_filename("1.bed"))
        self.history_panel_wait_for_hid_ok(1, allowed_force_refreshes=1)
        self.perform_upload(self.get_filename("2.bed"))
        self.history_panel_wait_for_hid_ok(2, allowed_force_refreshes=1)

        self.name = self._get_random_name(prefix="testcontents")

        self.libraries_open()
        self.libraries_index_create(self.name)
        self.libraries_open_with_name(self.name)

        self.libraries_dataset_import_from_history()
        self.libraries_dataset_import_from_history_select(["1.bed", "2.bed"])
        self.sleep_for(self.wait_types.UX_RENDER)
        self.libraries_dataset_import_from_history_click_ok()

        self.home()

        self.history_panel_create_new_with_name("new_history_for_library_list")
        self.libraries_open_with_name(self.name)
        self.sleep_for(self.wait_types.UX_RENDER)
        self.components.libraries.folder.select_all.wait_for_and_click()
        self.components.libraries.folder.add_to_history.wait_for_and_click()
        self.screenshot("libraries_add_to_history_menu")
        self.components.libraries.folder.add_to_history_collection.wait_for_and_click()
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("libraries_to_collection_landing")
        self.components.libraries.folder.import_datasets_ok_button.wait_for_and_click()

        self.collection_builder_set_name("my cool list")
        self.collection_builder_create()
        self.home()
        self.history_panel_wait_for_hid_ok(3)



from .framework import (
    selenium_test,
    SeleniumTestCase,
    UsesHistoryItemAssertions,
)

BUTTON_TOOLTIPS = {
    "display": 'View data',
    "edit": 'Edit attributes',
    "delete": 'Delete',
    "download": 'Download',
    "info": 'View details',
    "rerun": 'Run this job again',
}
EXPECTED_TOOLHELP_TITLE_TEXT = 'Tool help for Upload File'
TEST_DBKEY_TEXT = 'Honeybee (Apis mellifera): apiMel3 (apiMel3)'


class HistoryDatasetStateTestCase(SeleniumTestCase, UsesHistoryItemAssertions):
    hid = 1

    @selenium_test
    def test_dataset_state(self, hid=hid):
        item = self._prepare_dataset(self.hid)
        self.history_panel_item_body_component(hid, wait=True)

        self.assert_item_summary_includes(hid, "1 sequence")
        self.assert_item_dbkey_displayed_as(hid, "?")
        self.assert_item_info_includes(hid, 'uploaded fasta file')
        self.assert_item_peek_includes(hid, ">hg17")

        item.dbkey_button.wait_for_and_click()
        toolhelp_title_text = item.toolhelp_title.wait_for_visible().text
        # assert tool helptext
        assert EXPECTED_TOOLHELP_TITLE_TEXT == toolhelp_title_text, "Toolhelp title [%s] was not expected text [%s]." % (
            EXPECTED_TOOLHELP_TITLE_TEXT, toolhelp_title_text)

        self.screenshot("history_panel_dataset_expanded")

        self._assert_action_buttons(hid)

    @selenium_test
    def test_dataset_change_dbkey(self, hid=hid):
        item = self._prepare_dataset(hid)
        self.assert_item_dbkey_displayed_as(hid, "?")
        item.dbkey.wait_for_and_click()
        self.components.edit_dataset_attributes.database_build_dropdown.wait_for_and_click()
        # choose database option from 'Database/Build' dropdown, that equals to dbkey_text
        self.components.edit_dataset_attributes.dbkey_dropdown_results.dbkey_dropdown_option(
            dbkey_text=TEST_DBKEY_TEXT).wait_for_and_click()
        self.components.edit_dataset_attributes.save_btn.wait_for_and_click()
        self.history_panel_wait_for_hid_ok(hid)
        self.assert_item_dbkey_displayed_as(hid, "apiMel3")

    def _prepare_dataset(self, hid):
        self.register()
        self.perform_upload(self.get_filename("1.fasta"))
        self.history_panel_wait_for_hid_ok(hid)
        self.assert_item_name(hid, "1.fasta")
        self.assert_item_hid_text(hid)
        self._assert_title_buttons(hid)

        # Expand HDA and wait for details to show up.
        return self.history_panel_click_item_title(hid=hid, wait=True)

    def _assert_title_buttons(self, hid, expected_buttons=['display', 'edit', 'delete']):
        self._assert_buttons(hid, expected_buttons)

    def _assert_action_buttons(self, hid, expected_buttons=["info", "download"]):
        self._assert_buttons(hid, expected_buttons)

    def _assert_buttons(self, hid, expected_buttons):
        item_button = self.history_panel_item_component(hid=hid)
        for i, expected_button in enumerate(expected_buttons):

            # ensure old tooltip expired,
            # no tooltip appeared before the 1st element
            if i > 0:
                previous_button = item_button["%s_button" % expected_buttons[i - 1]].wait_for_visible()
                if previous_button.get_attribute("aria-describedby") is not None:
                    # wait for tooltip to disappear
                    self.components._.tooltip_balloon.wait_for_absent()

            button = item_button["%s_button" % expected_button]
            self.assert_tooltip_text(button.wait_for_visible(), BUTTON_TOOLTIPS[expected_button])



import json

from galaxy_test.base.workflow_fixtures import (
    WORKFLOW_NESTED_SIMPLE,
    WORKFLOW_SIMPLE_CAT_TWICE,
    WORKFLOW_SIMPLE_MAPPING,
    WORKFLOW_WITH_INVALID_STATE,
    WORKFLOW_WITH_OLD_TOOL_VERSION,
    WORKFLOW_WITH_OUTPUT_COLLECTION,
    WORKFLOW_WITH_RULES_1,
    WORKFLOW_WITH_RULES_2,
)
from .framework import (
    retry_assertion_during_transitions,
    retry_during_transitions,
    selenium_test,
    SeleniumTestCase
)


class WorkflowEditorTestCase(SeleniumTestCase):

    ensure_registered = True

    @selenium_test
    def test_basics(self):
        editor = self.components.workflow_editor
        annotation = "basic_test"
        name = self.workflow_create_new(annotation=annotation)
        edit_name_element = self.components.workflow_editor.edit_name.wait_for_visible()
        actual_name = edit_name_element.get_attribute("value")
        assert name in actual_name, "'{}' unequal name '{}'".format(name, actual_name)
        edit_annotation_element = self.components.workflow_editor.edit_annotation.wait_for_visible()
        actual_annotation = edit_annotation_element.get_attribute("value")
        assert annotation in actual_annotation, "'{}' unequal annotation '{}'".format(annotation, actual_annotation)

        editor.canvas_body.wait_for_visible()
        editor.tool_menu.wait_for_visible()

        self.screenshot("workflow_editor_blank")

        self.components._.left_panel_drag.wait_for_visible()
        self.components._.left_panel_collapse.wait_for_and_click()

        self.sleep_for(self.wait_types.UX_RENDER)

        self.screenshot("workflow_editor_left_collapsed")

        self.components._.right_panel_drag.wait_for_visible()
        self.components._.right_panel_collapse.wait_for_and_click()

        self.sleep_for(self.wait_types.UX_RENDER)

        self.screenshot("workflow_editor_left_and_right_collapsed")

    @selenium_test
    def test_data_input(self):
        editor = self.components.workflow_editor

        name = self.workflow_create_new()
        self.workflow_editor_add_input(item_name="data_input")
        self.screenshot("workflow_editor_data_input_new")
        editor.label_input.wait_for_and_send_keys("input1")
        editor.annotation_input.wait_for_and_send_keys("my cool annotation")
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_data_input_filled_in PRECLICK")
        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_data_input_filled_in")
        self.workflow_editor_click_save()

        self.workflow_index_open_with_name(name)
        data_input_node = editor.node._(label="input1")
        data_input_node.title.wait_for_and_click()
        label = editor.label_input.wait_for_value()
        assert label == "input1", label
        annotation = editor.annotation_input.wait_for_value()
        assert annotation == "my cool annotation"
        data_input_node.destroy.wait_for_and_click()
        data_input_node.wait_for_absent()
        self.screenshot("workflow_editor_data_input_deleted")

    @selenium_test
    def test_collection_input(self):
        editor = self.components.workflow_editor

        name = self.workflow_create_new()
        self.workflow_editor_add_input(item_name="data_collection_input")
        self.screenshot("workflow_editor_data_collection_input_new")
        editor.label_input.wait_for_and_send_keys("input1")
        editor.annotation_input.wait_for_and_send_keys("my cool annotation")
        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_data_collection_input_filled_in")
        self.workflow_editor_click_save()

        self.workflow_index_open_with_name(name)
        data_input_node = editor.node._(label="input1")
        data_input_node.title.wait_for_and_click()
        label = editor.label_input.wait_for_value()
        assert label == "input1", label
        annotation = editor.annotation_input.wait_for_value()
        assert annotation == "my cool annotation"
        data_input_node.destroy.wait_for_and_click()
        data_input_node.wait_for_absent()
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_data_collection_input_deleted")

    @selenium_test
    def test_integer_input(self):
        editor = self.components.workflow_editor

        name = self.workflow_create_new()
        self.workflow_editor_add_input(item_name="parameter_input")
        self.screenshot("workflow_editor_parameter_input_new")
        editor.label_input.wait_for_and_send_keys("input1")
        editor.annotation_input.wait_for_and_send_keys("my cool annotation")
        editor.label_input.wait_for_and_click()  # Seems to help force the save of whole annotation.
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_parameter_input_filled_in")
        self.workflow_editor_click_save()

        self.workflow_index_open_with_name(name)
        data_input_node = editor.node._(label="input1")
        data_input_node.title.wait_for_and_click()
        label = editor.label_input.wait_for_value()
        assert label == "input1", label
        annotation = editor.annotation_input.wait_for_value()
        assert annotation == "my cool annotation"
        data_input_node.destroy.wait_for_and_click()
        data_input_node.wait_for_absent()
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_parameter_input_deleted")

    @selenium_test
    def test_non_data_connections(self):
        self.open_in_workflow_editor("""
class: GalaxyWorkflow
inputs:
  input_int: integer
steps:
  simple_constructs:
    tool_id: simple_constructs
    label: tool_exec
    in:
      inttest: input_int
""")
        self.screenshot("workflow_editor_parameter_connection_simple")
        self.assert_connected("input_int#output", "simple_constructs#inttest")

        editor = self.components.workflow_editor

        tool_node = editor.node._(label="simple_constructs")
        tool_input = tool_node.input_terminal(name="inttest")
        tool_input.wait_for_and_click()

        editor.connector_destroy_callout.wait_for_and_click()
        self.assert_not_connected("input_int#output", "simple_constructs#inttest")
        self.screenshot("workflow_editor_parameter_connection_destroyed")

        # When connected, cannot turn it into a RuntimeValue..
        collapse_input = editor.collapse_icon(name="inttest")
        collapse_input.wait_for_absent_or_hidden()

        # If it is disconnected, then can specify as RuntimeValue
        connect_icon = editor.connect_icon(name="inttest")
        connect_icon.wait_for_visible()
        connect_icon.wait_for_and_click()
        collapse_input.wait_for_visible()

        # Also the connector should disappear
        tool_input.wait_for_absent_or_hidden()

        # Now make it connected again and watch the requestss
        connect_icon.wait_for_and_click()

        tool_input.wait_for_visible()
        collapse_input.wait_for_absent_or_hidden()

        self.workflow_editor_connect("input_int#output", "simple_constructs#inttest", screenshot_partial="workflow_editor_parameter_connection_dragging")
        self.assert_connected("input_int#output", "simple_constructs#inttest")

    @selenium_test
    def test_existing_connections(self):
        self.open_in_workflow_editor(WORKFLOW_SIMPLE_CAT_TWICE)

        editor = self.components.workflow_editor
        self.assert_connected("input1#output", "first_cat#input1")
        self.screenshot("workflow_editor_connection_simple")

        cat_node = editor.node._(label="first_cat")
        cat_input = cat_node.input_terminal(name="input1")
        cat_input.wait_for_and_click()
        editor.connector_destroy_callout.wait_for_visible()
        self.screenshot("workflow_editor_connection_callout")
        editor.connector_destroy_callout.wait_for_and_click()
        self.assert_not_connected("input1#output", "first_cat#input1")
        self.screenshot("workflow_editor_connection_destroyed")

        self.workflow_editor_connect("input1#output", "first_cat#input1", screenshot_partial="workflow_editor_connection_dragging")
        self.assert_connected("input1#output", "first_cat#input1")

    @selenium_test
    def test_rendering_output_collection_connections(self):
        self.open_in_workflow_editor(WORKFLOW_WITH_OUTPUT_COLLECTION)
        self.workflow_editor_maximize_center_pane()
        self.screenshot("workflow_editor_output_collections")

    @selenium_test
    def test_simple_mapping_connections(self):
        self.open_in_workflow_editor(WORKFLOW_SIMPLE_MAPPING)
        self.workflow_editor_maximize_center_pane()
        self.screenshot("workflow_editor_simple_mapping")
        self.assert_connected("input1#output", "cat#input1")
        self.assert_input_mapped("cat#input1")
        self.workflow_editor_destroy_connection("cat#input1")
        self.assert_input_not_mapped("cat#input1")
        self.assert_not_connected("input1#output", "cat#input1")
        self.workflow_editor_connect("input1#output", "cat#input1")
        self.assert_input_mapped("cat#input1")

    @selenium_test
    def test_rendering_simple_nested_workflow(self):
        self.open_in_workflow_editor(WORKFLOW_NESTED_SIMPLE)
        self.workflow_editor_maximize_center_pane()
        self.screenshot("workflow_editor_simple_nested")

    @selenium_test
    def test_rendering_rules_workflow_1(self):
        self.open_in_workflow_editor(WORKFLOW_WITH_RULES_1)
        rule_output = "apply#output"
        random_lines_input = "random_lines#input"
        self.workflow_editor_maximize_center_pane()
        self.screenshot("workflow_editor_rules_1")
        self.assert_connected(rule_output, random_lines_input)
        self.assert_input_mapped(random_lines_input)
        self.workflow_editor_destroy_connection(random_lines_input)
        self.assert_not_connected(rule_output, random_lines_input)
        self.assert_input_not_mapped(random_lines_input)
        self.workflow_editor_connect(rule_output, random_lines_input)
        self.assert_connected(rule_output, random_lines_input)
        self.assert_input_mapped(random_lines_input)

    @selenium_test
    def test_rendering_rules_workflow_2(self):
        self.open_in_workflow_editor(WORKFLOW_WITH_RULES_2)
        self.workflow_editor_maximize_center_pane(collapse_right=False)

        editor = self.components.workflow_editor
        rule_builder = self.components.rule_builder

        rule_output = "apply#output"
        copy_list_input = "copy_list#input1"

        apply_node = editor.node._(label="apply")

        self.assert_connected(rule_output, copy_list_input)
        self.assert_input_mapped(copy_list_input)
        self.workflow_editor_destroy_connection(copy_list_input)
        self.assert_not_connected(rule_output, copy_list_input)
        self.assert_input_not_mapped(copy_list_input)
        self.workflow_editor_connect(rule_output, copy_list_input)
        self.assert_connected(rule_output, copy_list_input)
        self.assert_input_mapped(copy_list_input)

        apply_node.title.wait_for_and_click()
        self.screenshot("workflow_editor_rules_2_form")
        self.tool_parameter_edit_rules()
        rule_builder._.wait_for_visible()
        self.screenshot("workflow_editor_rules_2_builder")
        new_rules = dict(
            rules=[{"type": "add_column_metadata", "value": "identifier0"}],
            mapping=[{"type": "list_identifiers", "columns": [0]}],
        )
        self.rule_builder_set_source(json.dumps(new_rules))
        rule_builder.main_button_ok.wait_for_and_click()
        apply_node.title.wait_for_and_click()
        self.sleep_for(self.wait_types.UX_RENDER)
        # screenshot should have async warning about connection removed
        self.screenshot("workflow_editor_rules_2_after_change")
        self.assert_input_not_mapped(copy_list_input)
        # changing output collection type remove outbound connections, so this
        # this needs to be re-connected. Remove this re-connection if we upgrade
        # the workflow editor to try to re-establish the connection with different
        # mapping.
        self.workflow_editor_connect(rule_output, copy_list_input)
        self.assert_connected(rule_output, copy_list_input)
        # Regardless - this rules now say to connect a list to a list instead of a list
        # to a list:list, so there should be no mapping anymore even after connected.
        self.assert_input_not_mapped(copy_list_input)

    @selenium_test
    def test_save_as(self):
        name = self.workflow_upload_yaml_with_random_name(WORKFLOW_SIMPLE_CAT_TWICE)
        self.workflow_index_open()
        self.workflow_index_open_with_name(name)
        self.sleep_for(self.wait_types.UX_RENDER)
        self.screenshot("workflow_editor_edit_menu")
        self.workflow_editor_click_option("Save As")

    @selenium_test
    def test_editor_tool_upgrade_message(self):
        workflow_populator = self.workflow_populator
        workflow_populator.upload_yaml_workflow(WORKFLOW_WITH_OLD_TOOL_VERSION, exact_tools=True)
        self.workflow_index_open()
        self.workflow_index_click_option("Edit")
        self.assert_modal_has_text("Using version '0.2' instead of version '0.0.1'")
        self.screenshot("workflow_editor_tool_upgrade")

    @selenium_test
    def test_editor_invalid_tool_state(self):
        workflow_populator = self.workflow_populator
        workflow_populator.upload_yaml_workflow(WORKFLOW_WITH_INVALID_STATE, exact_tools=True)
        self.workflow_index_open()
        self.workflow_index_click_option("Edit")
        self.assert_modal_has_text("Using version '0.2' instead of version '0.0.1'")
        self.assert_modal_has_text("Using default: '1'")
        self.screenshot("workflow_editor_invalid_state")

    @selenium_test
    def test_missing_tools(self):
        workflow_populator = self.workflow_populator
        workflow_populator.upload_yaml_workflow("""
class: GalaxyWorkflow
inputs:
  - id: input1
steps:
  - tool_id: missing
    label: first_cat
    state:
      foo: bar
""")
        self.workflow_index_open()
        self.workflow_index_click_option("Edit")
        self.assert_modal_has_text("Tool is not installed")
        self.screenshot("workflow_editor_missing_tool")

    @selenium_test
    def test_workflow_bookmarking(self):
        @retry_during_transitions
        def assert_workflow_bookmarked_status(target_status):
            name_matches = [c.text == new_workflow_name for c in self.components.tool_panel.workflow_names.all()]
            status = any(name_matches)
            self.assertTrue(status == target_status)

        new_workflow_name = self.workflow_create_new(clear_placeholder=True)

        # Assert workflow not initially bookmarked.
        assert_workflow_bookmarked_status(False)

        save_button = self.components.workflows.save_button
        save_button.wait_for_clickable()

        # element is clickable, but still might be behind the modal
        self.wait_for_selector_absent_or_hidden(self.modal_body_selector())
        save_button.wait_for_and_click()

        # wait for saving
        self.wait_for_selector_absent_or_hidden(self.modal_body_selector())
        self.driver.find_element_by_id("workflow").click()

        # parse workflow table
        table_elements = self.workflow_index_table_elements()
        self.sleep_for(self.wait_types.UX_RENDER)
        bookmark_td = table_elements[0].find_elements_by_tag_name('td')[4]

        # get bookmark pseudo element
        # https://stackoverflow.com/questions/45427223/click-on-pseudo-element-using-selenium
        self.action_chains().move_to_element_with_offset(bookmark_td, 20, 20).click().perform()
        self.sleep_for(self.wait_types.UX_TRANSITION)

        # search for bookmark in tools menu
        self.components.tool_panel.search.wait_for_and_send_keys(new_workflow_name)
        assert_workflow_bookmarked_status(True)

    def workflow_editor_maximize_center_pane(self, collapse_left=True, collapse_right=True):
        if collapse_left:
            self.components._.left_panel_collapse.wait_for_and_click()
        if collapse_right:
            self.components._.right_panel_collapse.wait_for_and_click()
        self.sleep_for(self.wait_types.UX_RENDER)

    def workflow_editor_connect(self, source, sink, screenshot_partial=None):
        source_id, sink_id = self.workflow_editor_source_sink_terminal_ids(source, sink)
        source_element = self.driver.find_element_by_css_selector("#" + source_id)
        sink_element = self.driver.find_element_by_css_selector("#" + sink_id)

        ac = self.action_chains()
        ac = ac.move_to_element(source_element).click_and_hold()
        if screenshot_partial:
            ac = ac.move_to_element_with_offset(sink_element, -5, 0)
            ac.perform()
            self.sleep_for(self.wait_types.UX_RENDER)
            self.screenshot(screenshot_partial)
            ac = self.action_chains()

        ac = ac.move_to_element(sink_element).release().perform()

    def assert_connected(self, source, sink):
        source_id, sink_id = self.workflow_editor_source_sink_terminal_ids(source, sink)
        self.components.workflow_editor.connector_for(source_id=source_id, sink_id=sink_id).wait_for_visible()

    def assert_not_connected(self, source, sink):
        source_id, sink_id = self.workflow_editor_source_sink_terminal_ids(source, sink)
        self.components.workflow_editor.connector_for(source_id=source_id, sink_id=sink_id).wait_for_absent()

    def open_in_workflow_editor(self, yaml_content):
        name = self.workflow_upload_yaml_with_random_name(yaml_content)
        self.workflow_index_open()
        self.workflow_index_open_with_name(name)
        self.workflow_editor_click_option("Auto Layout")

    def workflow_editor_source_sink_terminal_ids(self, source, sink):
        editor = self.components.workflow_editor

        source_node_label, source_output = source.split("#", 1)
        sink_node_label, sink_input = sink.split("#", 1)

        source_node = editor.node._(label=source_node_label)
        sink_node = editor.node._(label=sink_node_label)

        source_node.wait_for_visible()
        sink_node.wait_for_visible()

        output_terminal = source_node.output_terminal(name=source_output)
        input_terminal = sink_node.input_terminal(name=sink_input)

        output_element = output_terminal.wait_for_visible()
        input_element = input_terminal.wait_for_visible()

        source_id = output_element.get_attribute("id")
        sink_id = input_element.get_attribute("id")

        return source_id, sink_id

    def workflow_editor_add_input(self, item_name="data_input"):
        editor = self.components.workflow_editor

        # Make sure we're on the the workflow editor and not clicking the main tool panel.
        editor.canvas_body.wait_for_visible()

        editor.tool_menu.wait_for_visible()
        editor.tool_menu_section_link(section_name="inputs").wait_for_and_click()
        editor.tool_menu_item_link(item_name=item_name).wait_for_and_click()

    def workflow_editor_destroy_connection(self, sink):
        editor = self.components.workflow_editor

        sink_node_label, sink_input_name = sink.split("#", 1)
        sink_node = editor.node._(label=sink_node_label)
        sink_input = sink_node.input_terminal(name=sink_input_name)
        sink_input.wait_for_and_click()
        editor.connector_destroy_callout.wait_for_and_click()

    def assert_input_mapped(self, sink):
        editor = self.components.workflow_editor
        sink_node_label, sink_input_name = sink.split("#", 1)
        sink_node = editor.node._(label=sink_node_label)
        sink_mapping_icon = sink_node.input_mapping_icon(name=sink_input_name)
        sink_mapping_icon.wait_for_visible()

    def assert_input_not_mapped(self, sink):
        editor = self.components.workflow_editor
        sink_node_label, sink_input_name = sink.split("#", 1)
        sink_node = editor.node._(label=sink_node_label)
        sink_mapping_icon = sink_node.input_mapping_icon(name=sink_input_name)
        sink_mapping_icon.wait_for_absent_or_hidden()

    def workflow_index_open_with_name(self, name):
        self.workflow_index_open()
        self.workflow_index_search_for(name)
        self.workflow_index_click_option("Edit")

    def workflow_upload_yaml_with_random_name(self, content):
        workflow_populator = self.workflow_populator
        name = self._get_random_name()
        workflow_populator.upload_yaml_workflow(content, name=name)
        return name

    def workflow_create_new(self, annotation=None, clear_placeholder=False):
        self.workflow_index_open()
        self.sleep_for(self.wait_types.UX_RENDER)
        self.click_button_new_workflow()
        self.sleep_for(self.wait_types.UX_RENDER)
        form_element = self.driver.find_element_by_id("submit")
        name = self._get_random_name()
        annotation = annotation or self._get_random_name()
        inputs = self.driver.find_elements_by_class_name("ui-input")
        if clear_placeholder:
            inputs[0].clear()
        inputs[0].send_keys(name)
        inputs[1].send_keys(annotation)
        form_element.click()
        return name

    @retry_assertion_during_transitions
    def assert_modal_has_text(self, expected_text):
        modal_element = self.wait_for_selector_visible(self.modal_body_selector())
        text = modal_element.text
        assert expected_text in text, "Failed to find expected text [{}] in modal text [{}]".format(expected_text, text)
